#include <ncbi.h>
#include <dfa.h>

/*************************************************************************
 * dfa_size(dp)
 *
 *	Return the no. of bytes of storage in use by the DFA.
 *	This number may be considerably less than the no. of bytes
 *	allocated to the DFA, as reported by dfa_extent().
 *
 *	Returns 0 upon error, and sets dfaerrno = the error encountered.
 *************************************************************************/
unsigned long _cdecl
dfa_size(dp)
	register DFAPtr	dp;
{
	register DFA_BufPtr	bp;
	register long	totbytes = 0;

	if (dp == NULL || dp->magic != DFA_MAGIC) {
		dfaerrno = dfaErrBadPtr;
		return 0;
	}

	if (dp->opstate < dfaOpstateMemInit || dp->opstate >= dfaOpstateDestroyed) {
		dfaerrno = dfaErrOpstate;
		return 0;
	}

	bp = &dp->buf0;
	while (bp != NULL) {
		if (bp->chain != NULL)
			totbytes += bp->bused;
		else
			totbytes += bp->bsize - dp->memavail;
		bp = bp->chain;
	}
	return totbytes;
}
