#include <ncbi.h>
#include <gishlib.h>

/*
	fct_factorial(n)

	Returns n! as a double.

	Author:  Warren Gish, Washington University School of Medicine, St. Louis
*/

double
fct_factorial(n)
	long	n;
{
	static double	precomputed[100]
		= { 1., 1., 2., 6., 24., 120., 720., 5040., 40320., 362880., 3628800. };
	static volatile sig_atomic_t	filled;

	if (n < 0)
		return 0.0; /* Undefined! */

	if (n >= DIM(precomputed))
		return floor(0.5 + fct_gamma((double)(n+1)));

	if (!filled) {
		mproc_lock();
		if (!filled) {
			register int	m;

			for (m = 10; m < DIM(precomputed)-1; ++m) {
				precomputed[m+1] = precomputed[m] * (m+1);
			}
			filled = 1;
		}
		mproc_unlock();
	}
	return precomputed[n];
}

